package ru.dbotthepony.mc.otm.block.entity.blackhole

import it.unimi.dsi.fastutil.objects.ObjectArrayList
import it.unimi.dsi.fastutil.objects.ObjectArraySet
import it.unimi.dsi.fastutil.objects.ObjectIterators
import net.minecraft.core.BlockPos
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.level.block.state.BlockState
import net.neoforged.neoforge.capabilities.Capabilities
import ru.dbotthepony.kommons.util.DelegateSetter
import ru.dbotthepony.kommons.util.getValue
import ru.dbotthepony.kommons.util.setValue
import ru.dbotthepony.kommons.util.value
import ru.dbotthepony.mc.otm.block.entity.MatteryDeviceBlockEntity
import ru.dbotthepony.mc.otm.block.entity.tech.EnergyHatchBlockEntity
import ru.dbotthepony.mc.otm.block.entity.tech.MatterHatchBlockEntity
import ru.dbotthepony.mc.otm.capability.FlowDirection
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.energy.CombinedProfiledEnergyStorage
import ru.dbotthepony.mc.otm.capability.energy.receiveEnergy
import ru.dbotthepony.mc.otm.capability.matter.CombinedProfiledMatterStorage
import ru.dbotthepony.mc.otm.config.MachinesConfig
import ru.dbotthepony.mc.otm.core.RandomSource2Generator
import ru.dbotthepony.mc.otm.core.collect.map
import ru.dbotthepony.mc.otm.core.multiblock.ShapedMultiblock
import ru.dbotthepony.mc.otm.core.getBlockStateNow
import ru.dbotthepony.mc.otm.core.immutableList
import ru.dbotthepony.mc.otm.core.math.Decimal
import ru.dbotthepony.mc.otm.core.math.plus
import ru.dbotthepony.mc.otm.core.math.times
import ru.dbotthepony.mc.otm.core.multiblock.BlockEntityTag
import ru.dbotthepony.mc.otm.core.multiblock.Strategy
import ru.dbotthepony.mc.otm.core.multiblock.shapedMultiblock
import ru.dbotthepony.mc.otm.menu.tech.BlackHoleGeneratorMenu
import ru.dbotthepony.mc.otm.registry.MBlockEntities
import ru.dbotthepony.mc.otm.registry.MBlockTags
import ru.dbotthepony.mc.otm.registry.MBlocks
import java.io.Closeable

class BlackHoleGeneratorBlockEntity(blockPos: BlockPos, blockState: BlockState) : MatteryDeviceBlockEntity(MBlockEntities.BLACK_HOLE_GENERATOR, blockPos, blockState) {
	var multiblock: ShapedMultiblock? = null
		private set
	private var multiblockSyncher: Closeable? = null

	var lastRange by syncher.int(-1, setter = DelegateSetter { field, value ->
		if (value == -1) {
			multiblockSyncher?.close()
			multiblockSyncher = null
			multiblock?.close()
			multiblock = null
			field.accept(value)
		} else if (value != field.value) {
			multiblockSyncher?.close()
			multiblock = CONFIGURATIONS[value - 5].value.create(blockPos)
			multiblockSyncher = syncher.add0(multiblock!!)
			field.accept(value)
		}
	})
		private set

	var drawBuildingGuide by syncher.boolean()

	val energy = CombinedProfiledEnergyStorage(
		FlowDirection.NONE,
		{ multiblock?.blockEntities(EnergyHatchBlockEntity.OUTPUT_TAG)?.iterator()?.map { it.energy } ?: ObjectIterators.emptyIterator() },
		{ level?.random?.let { RandomSource2Generator(it) } }
	)

	val matter = CombinedProfiledMatterStorage(
		FlowDirection.NONE,
		{ multiblock?.blockEntities(MatterHatchBlockEntity.INPUT_TAG)?.iterator()?.map { it.matter } ?: ObjectIterators.emptyIterator() },
		{ level?.random?.let { RandomSource2Generator(it) } }
	)

	init {
		exposeSideless(MatteryCapability.MATTER_BLOCK, matter)
		exposeSideless(MatteryCapability.BLOCK_ENERGY, energy)
		exposeSideless(Capabilities.EnergyStorage.BLOCK, energy)
	}

	private fun findBlackHoleRange(): Int {
		val normal = blockRotation.normal
		val level = level!!

		if (lastRange != -1) {
			val pos = blockPos + normal * lastRange

			if (level.getBlockStateNow(pos).block == MBlocks.BLACK_HOLE) {
				return lastRange
			}
		}

		for (i in 5 until 64) {
			val pos = blockPos + normal * i

			if (level.getBlockStateNow(pos).block == MBlocks.BLACK_HOLE) {
				return i
			}
		}

		return -1
	}

	override fun tick() {
		super.tick()

		energy.tick()
		matter.tick()

		val level = level!!

		if (lastRange == -1) {
			this.lastRange = findBlackHoleRange()
		}

		val multiblock = multiblock

		if (multiblock != null) {
			if (!multiblock.update(level, blockRotation.front)) {
				this.lastRange = findBlackHoleRange()
			} else {
				val blackHole = multiblock.blockEntities(BLACK_HOLE).first()

				if (matter.extractMatter(MachinesConfig.BlackHoleGenerator.MATTER_RATE, true) != MachinesConfig.BlackHoleGenerator.MATTER_RATE) return
				val energyToStore = blackHole.mass / MachinesConfig.BlackHoleGenerator.MASS_DIVISOR
				if (energy.receiveEnergy(energyToStore, true) != energyToStore) return

				matter.extractMatter(MachinesConfig.BlackHoleGenerator.MATTER_RATE, false)
				energy.receiveEnergy(energyToStore, false)

				blackHole.mass += MachinesConfig.BlackHoleGenerator.MASS_FEEDING_RATIO * MachinesConfig.BlackHoleGenerator.MATTER_RATE
			}
		}
	}

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu? {
		return BlackHoleGeneratorMenu(containerID, inventory, this)
	}

	companion object {
		val BLACK_HOLE = BlockEntityTag(BlackHoleBlockEntity::class)

		private fun points(x: Int, z: Int, result: ObjectArraySet<BlockPos>) {
			if (x == 0) {
				result.add(BlockPos(0, 0, z))
				result.add(BlockPos(0, 0, -z))
				result.add(BlockPos(z, 0, 0))
				result.add(BlockPos(-z, 0, 0))
			} else if (x == z) {
				result.add(BlockPos(x, 0, z))
				result.add(BlockPos(-x, 0, z))
				result.add(BlockPos(x, 0, -z))
				result.add(BlockPos(-x, 0, -z))
			} else if (x < z) {
				result.add(BlockPos(x, 0, z))
				result.add(BlockPos(-x, 0, z))
				result.add(BlockPos(x, 0, -z))
				result.add(BlockPos(-x, 0, -z))

				result.add(BlockPos(z, 0, x))
				result.add(BlockPos(-z, 0, x))
				result.add(BlockPos(z, 0, -x))
				result.add(BlockPos(-z, 0, -x))
			}
		}

		private fun ring(size: Int): ObjectArrayList<BlockPos> {
			val positions = ObjectArraySet<BlockPos>()
			var x = 0
			var y = size
			var p = (5 - size * 4) / 4
			points(x, y, positions)

			while (x < y) {
				x++

				if (p < 0) {
					p += x * 2 + 1
				} else {
					y--
					p += (x - y) * 2 + 1
				}

				points(x, y, positions)
			}

			val result = ObjectArrayList<BlockPos>(positions.size)

			for (pos in positions) {
				if (pos.x == 0 && pos.z + size == 0) continue
				result.add(BlockPos(pos.x, 0, pos.z + size))
			}

			return result
		}

		val CONFIGURATIONS = immutableList {
			for (i2 in 0 until 59) {
				val i = i2 + 5

				accept(lazy {
					shapedMultiblock {
						builder.customCheck {
							val blocks = it.blocks()

							blocks.getInt(MBlocks.MATTER_INJECTOR) == 1 &&
							blocks.getInt(MBlocks.ANTIMATTER_INJECTOR) == 1 &&
							blocks.getInt(MBlocks.HIGH_ENERGY_PARTICLE_COLLECTOR) == 1
						}

						block(MBlocks.BLACK_HOLE_GENERATOR)

						relative(BlockPos(0, 0, i)) {
							block(MBlocks.BLACK_HOLE)
							tag(BLACK_HOLE)
						}

						val ring = ring(i)

						for (pos in ring) {
							val node = relative(pos)

							if ((pos.x == i || pos.x == -i) && pos.z == i) {
								node.block(MBlocks.MATTER_INJECTOR)
								node.block(MBlocks.ANTIMATTER_INJECTOR)
								node.tagBlock()
							} else if (pos.x == 0 && pos.z == i * 2) {
								node.block(MBlocks.HIGH_ENERGY_PARTICLE_COLLECTOR)
								node.tagBlock()
							} else if (pos.x == i || pos.x == -i || pos.z == i * 2 || pos.z == 0) {
								node.or {
									block(MBlocks.ENERGY_OUTPUT_HATCH)
									tag(EnergyHatchBlockEntity.OUTPUT_TAG)
								}

								node.or {
									block(MBlocks.MATTER_INPUT_HATCH)
									tag(MatterHatchBlockEntity.INPUT_TAG)
								}

								node.block(MBlockTags.MULTIBLOCK_STRUCTURE)
							} else {
								node.block(MBlockTags.MULTIBLOCK_HARD_STRUCTURE)
							}
						}
					}
				})
			}
		}
	}
}
